{*******************************************************}
{                                                       }
{       Advanced SQL statement parser                   }
{       Copyright (c) 2001, 2002 AS Gaiasoft            }
{       Created by Gert Kello                           }
{                                                       }
{*******************************************************}

unit gaSQLParserHelperClasses;

interface

uses
  gaAdvancedSQLParser, gaBasicSQLParser;

type
  TgaSQLStatementPart = class (TgaSQLTokenList)
  private
    FIsInvalid: Boolean;
    FParseComplete: Boolean;
    function GetParseComplete: Boolean;
    procedure SetIsInvalid(const Value: Boolean);
    procedure SetParseComplete(const Value: Boolean);
  protected
    procedure CheckModifyAllowed; virtual;
    function GetAsString: string; override;
    function InternalGetParseComplete: Boolean; virtual;
    procedure InternalSetParseComplete; virtual;
    procedure SetAsString(const Value: string); virtual;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    property AsString read GetAsString write SetAsString;
    property IsInvalid: Boolean read FIsInvalid write SetIsInvalid;
    property ParseComplete: Boolean read GetParseComplete write 
            SetParseComplete;
  end;
  
  TgaSQLExpression = class (TgaSQLStatementPart)
  private
    FIsCorrect: Boolean;
    FIsExpression: Boolean;
    FOldOnStatementAdd: TgaTokenEvent;
    FParenCount: Integer;
    function GetCanParseEnd: Boolean;
    function GetExpression: string;
  protected
    procedure AddParen(AToken: TgaSQLTokenObj); virtual;
    procedure ParseExpression(AToken: TgaSQLTokenObj); virtual;
    procedure RemoveParen(AToken: TgaSQLTokenObj); virtual;
    procedure SetIsExpression(const Value: Boolean);
    property CanParseEnd: Boolean read GetCanParseEnd;
    property ParenCount: Integer read FParenCount write FParenCount;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    procedure ExecuteTokenAdded(Sender: TObject; AToken: TgaSQLTokenObj); 
            override;
    property Expression: string read GetExpression;
    property IsCorrect: Boolean read FIsCorrect;
    property IsExpression: Boolean read FIsExpression;
  end;
  
  TgaSQLStatementPartClass = class of TgaSQLStatementPart;

  TFieldExprParseState = (fepsNoSymbolsParsed, fepsFieldExpression,
    fepsFieldAlias, fepsExpression, fepsFieldAliasParsed, fepsFieldParseFinished);
  TgaSQLField = class (TgaSQLExpression)
  private
    FFieldAlias: TgaSQLTokenListBookmark;
    FFieldName: TgaSQLTokenListBookmark;
    FFieldParseState: TFieldExprParseState;
    FFieldPrefixies: TgaSQLTokenList;
    function GetFieldAlias: string;
    function GetFieldName: string;
    function GetFieldPrefix: string;
    procedure SetFieldAlias(const Value: string);
    procedure SetFieldName(const Value: string);
    procedure SetFieldPrefix(const Value: string);
  protected
    procedure InternalSetParseComplete; override;
    procedure ParseFieldAlias(AToken: TgaSQLTokenObj);
    procedure ParseFieldExpression(AToken: TgaSQLTokenObj);
    procedure ParsePreFieldExpression(AToken: TgaSQLTokenObj);
    procedure StartExpressionParse;
    procedure TerminateFieldExpressionParse;
    property FieldParseState: TFieldExprParseState read FFieldParseState;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    destructor Destroy; override;
    procedure ExecuteTokenAdded(Sender: TObject; AToken: TgaSQLTokenObj); 
            override;
    property FieldAlias: string read GetFieldAlias write SetFieldAlias;
    property FieldAliasToken: TgaSQLTokenListBookmark read FFieldAlias;
    property FieldName: string read GetFieldName write SetFieldName;
    property FieldNameToken: TgaSQLTokenListBookmark read FFieldName;
    property FieldPrefix: string read GetFieldPrefix write SetFieldPrefix;
    property FieldPrefixies: TgaSQLTokenList read FFieldPrefixies;
  end;
  
  TTableExprParseState = (tepsNoSymbolsParsed, tepsTableName, tepsTableAlias,
    tepsTableAliasParsed, tepsTableParseFinished);
  TgaSQLTable = class (TgaSQLStatementPart)
  private
    FIsAliasAllowed: Boolean;
    FTableAlias: TgaSQLTokenListBookmark;
    FTableName: TgaSQLTokenListBookmark;
    FTableParseState: TTableExprParseState;
    FTablePrefixies: TgaSQLTokenList;
    function GetTableAlias: string;
    function GetTableName: string;
    function GetTablePrefix: string;
    procedure SetTableAlias(const Value: string);
    procedure SetTableName(const Value: string);
    procedure SetTablePrefix(const Value: string);
  protected
    procedure InternalSetParseComplete; override;
    procedure ParsePreTableName(AToken: TgaSQLTokenObj);
    procedure ParseTableAlias(AToken: TgaSQLTokenObj);
    procedure ParseTableName(AToken: TgaSQLTokenObj);
    property TableParseState: TTableExprParseState read FTableParseState write 
            FTableParseState;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement); override;
    destructor Destroy; override;
    procedure ExecuteTokenAdded(Sender: TObject; AToken: TgaSQLTokenObj); 
            override;
    property IsAliasAllowed: Boolean read FIsAliasAllowed write FIsAliasAllowed;
    property TableAlias: string read GetTableAlias write SetTableAlias;
    property TableAliasToken: TgaSQLTokenListBookmark read FTableAlias;
    property TableName: string read GetTableName write SetTableName;
    property TableNameToken: TgaSQLTokenListBookmark read FTableName;
    property TablePrefix: string read GetTablePrefix write SetTablePrefix;
    property TablePrefixies: TgaSQLTokenList read FTablePrefixies;
  end;
  
  TgaSQLWhereExpression = class (TgaSQLExpression)
  protected
    function GetAsString: string; override;
    procedure SetAsString(const Value: string); override;
  end;
  
  TgaJoinOnClause = class (TgaSQLExpression)
  protected
    function GetAsString: string; override;
    procedure SetAsString(const Value: string); override;
  end;
  
  TgaSQLStatementPartList = class (TgaListOfSQLTokenLists)
  private
    FOwnerStatement: TgaCustomSQLStatement;
    FParseComplete: Boolean;
    FStatementPartType: TgaSQLStatementPartClass;
    FTokenList: TgaSQLTokenList;
    function GetParseComplete: Boolean;
    procedure SetParseComplete(const Value: Boolean);
  protected
    FCurrentPart: TgaSQLStatementPart;
    function GetAsString: string; override;
    function GetCurrentPart: TgaSQLStatementPart; virtual;
    procedure InternalSetParseComplete; virtual;
    procedure SetAsString(const Value: string); virtual;
    property OwnerStatement: TgaCustomSQLStatement read FOwnerStatement;
    property StatementPartType: TgaSQLStatementPartClass read 
            FStatementPartType;
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement; 
            AStatementPartType: TgaSQLStatementPartClass);
    destructor Destroy; override;
    procedure Clear; override;
    procedure ExecuteTokenAdded(Sender: TObject; AToken: TgaSQLTokenObj); 
            override;
    property AsString read GetAsString write SetAsString;
    property CurrentPart: TgaSQLStatementPart read GetCurrentPart;
    property ParseComplete: Boolean read GetParseComplete write 
            SetParseComplete;
    property TokenList: TgaSQLTokenList read FTokenList;
  end;
  
  TgaSQLFieldList = class (TgaSQLStatementPartList)
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement);
  end;
  
  TgaSQLTableList = class (TgaSQLStatementPartList)
  public
    constructor Create(AOwnerStatement: TgaCustomSQLStatement);
  end;
  
  TgaSQLOrderByList = class (TgaSQLFieldList)
  protected
    function GetAsString: string; override;
    procedure SetAsString(const Value: string); override;
  end;
  
procedure ParseStringToTokens(const AString: string; ATokenList: TgaSQLTokenHolderList);

implementation

uses
  SysUtils, TypInfo, gaSQLParserConsts;

procedure ParseStringToTokens(const AString: string; ATokenList: TgaSQLTokenHolderList);
var
  lSQLParser: TgaBasicSQLParser;
  lTokenObj: TgaSQLTokenObj;
begin
  ATokenList.Clear;
  lSQLParser := TgaBasicSQLParser.Create(nil);
  try
    lSQLParser.SQLText.Text := AString;
    while lSQLParser.TokenType <> stEnd do
    begin
      lTokenObj := ATokenList.NewToken;
      lTokenObj.AssignTokenInfo(lSQLParser);
      lSQLParser.NextToken;
    end;
  finally
    lSQLParser.Free;
  end;
end;

{
********************************* TgaSQLTable **********************************
}
constructor TgaSQLTable.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited;
  FIsAliasAllowed := True;
  OwnerStatement.AddTable(Self);
end;

destructor TgaSQLTable.Destroy;
begin
  OwnerStatement.RemoveTable(Self);
  FTableAlias.Free;
  FTableName.Free;
  FTablePrefixies.Free;
  inherited Destroy;
end;

procedure TgaSQLTable.ExecuteTokenAdded(Sender: TObject; AToken: 
        TgaSQLTokenObj);
begin
  inherited;
  if AToken.TokenType = stComment then
    Exit;
  case TableParseState of
    tepsNoSymbolsParsed:
      ParsePreTableName(AToken);
    tepsTableName:
      ParseTableName(AToken);
    tepsTableAlias:
      ParseTableAlias(AToken);
    tepsTableAliasParsed:
      if not (AToken.TokenType in [stDelimitier, stComma, stEnd]) then
        IsInvalid := True;
    tepsTableParseFinished:
      raise Exception.Create(SerrTableParseFinised);
    { fepsExpression:
      Shouldn't get here as IsExpression should be true then }
    else
      raise Exception.CreateFmt(SerrUnecpectedFieldParseState,
        [GetEnumName(TypeInfo(TTableExprParseState), Ord(TableParseState))]);
  end;
  if (AToken.TokenType = stComma) then
    ParseComplete := True;
end;

function TgaSQLTable.GetTableAlias: string;
begin
  if Assigned(FTableAlias) then
    Result := GetTokenObjAsString(FTableAlias.TokenObj)
  else
    Result := EmptyStr;
end;

function TgaSQLTable.GetTableName: string;
begin
  if Assigned(FTableName) then
    Result := GetTokenObjAsString(FTableName.TokenObj)
  else
    Result := EmptyStr;
end;

function TgaSQLTable.GetTablePrefix: string;
begin
  Result := TablePrefixies.AsString;
end;

procedure TgaSQLTable.InternalSetParseComplete;
begin
  inherited;
  TableParseState := tepsTableParseFinished;
end;

procedure TgaSQLTable.ParsePreTableName(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      FTablePrefixies := TgaSQLTokenList.CreateMirror(OwnerStatement, OwnerStatement.CurrentSQL);
      FTablePrefixies.SetStartPos(OwnerStatement.CurrentSQL, True);
      FTableParseState := tepsTableName;
    end;
    stDelimitier:
      {skip leading delimitiers};
    else
      IsInvalid := True;
  end;
end;

procedure TgaSQLTable.ParseTableAlias(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stDelimitier:
      {no special processing};
    stSymbol, stQuotedSymbol, stString:
      if not AToken.TokenSymbolIs('AS') then
      begin
        FTableAlias := GetBookmark;
        FTableParseState := tepsTableAliasParsed;
      end;
    stOther:
      if AToken.TokenString = ';' then
        FTableParseState := tepsTableAliasParsed
      else
        IsInvalid := True;
    else
      IsInvalid := True;
  end;
  
end;

procedure TgaSQLTable.ParseTableName(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
      ; { Do nothing as added tokens are inserted at the end of the prefixies
          list automatically }
    stDelimitier, stComma:
    begin
      FTablePrefixies.Last;
      // last one should be the current comma or delimitier
      FTablePrefixies.Previous;
      if FTablePrefixies.CurrentItem.TokenType <> stPeriod then
      begin
        Locate(FTablePrefixies.CurrentItem);
        FTableName := GetBookmark;
        FTablePrefixies.Previous;
      end;
      FTablePrefixies.SetEndPos(FTablePrefixies, True);
      if IsAliasAllowed then
        FTableParseState := tepsTableAlias
      else
        FTableParseState := tepsTableAliasParsed;
    end;
    stOther:
      if AToken.TokenString = ';' then
        FTableParseState := tepsTableAliasParsed
      else
        IsInvalid := True;
    else
      IsInvalid := True;
  end;
end;

procedure TgaSQLTable.SetTableAlias(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
  tmpToken: TgaSQLTokenObj;
begin
  if TableAlias <> Value then
  begin
    CheckModifyAllowed;
    if (not IsAliasAllowed) and (Value <> EmptyStr) then
      raise Exception.Create(SerrTableAliasNotAllowed);
    if not Assigned(FTableName) then
      raise Exception.Create(SerrNoTableForAlias);
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count > 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Table alias', 1, tmpTokenList.Count]);
      tmpTokenList.First;
      if Assigned(FTableAlias) then
        FTableAlias.TokenObj := tmpTokenList.CurrentItem
      else begin
        tmpToken := TgaSQLTokenObj.Create;
        tmpToken.SetTokenInfo(' ', stDelimitier, False, #0);
        GotoBookmark(FTableName);
        InsertAfterCurrent(tmpToken, True);
        InsertAfterCurrent(tmpTokenList.CurrentItem, True);
        FTableAlias := GetBookmark;
      end;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLTable.SetTableName(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
begin
  if TableName <> Value then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count <> 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Table name', 1, tmpTokenList.Count]);
      tmpTokenList.First;
      FTableName.TokenObj := tmpTokenList.CurrentItem;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLTable.SetTablePrefix(const Value: string);
var
  tmpStr: string;
  tmpTokenList: TgaSQLTokenHolderList;
begin
  if Value[Length(Value)] = '.' then
    tmpStr := Value
  else
    tmpStr := Value + '.';
  if (TablePrefix <> tmpStr) then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(tmpStr, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      TablePrefixies.CopyListContest(tmpTokenList);
    finally
      tmpTokenList.Free;
    end;
  end;
end;

{
**************************** TgaSQLWhereExpression *****************************
}
function TgaSQLWhereExpression.GetAsString: string;
begin
  Result := EmptyStr;
  if IsEmpty then
    Exit;
  First;
  if CurrentItem.TokenSymbolIs('WHERE') then
  begin
    Next;
    while (not Eof) and (CurrentItem.TokenType = stDelimitier) do
      Next;
  end;
  while not Eof do
  begin
    Result := Result + CurrentItem.TokenAsString;
    Next;
  end;
end;

procedure TgaSQLWhereExpression.SetAsString(const Value: string);
var
  tmpStr: string;
  tmpToken: TgaSQLTokenObj;
begin
  if (Trim(Value) = '') or (SameText('where ', Copy(Value, 1, 6))) then
    tmpStr := Value
  else
    tmpStr := 'where ' + Value;
  inherited SetAsString(tmpStr);
  Last;
  if CurrentItem.TokenType <> stDelimitier then
  begin
    tmpToken := TgaSQLTokenObj.Create;
    tmpToken.SetTokenInfo(#13#10, stDelimitier, False, #0);
    Add(tmpToken);
    ExecuteTokenAdded(Self, tmpToken);
  end;
end;

{
********************************* TgaSQLField **********************************
}
constructor TgaSQLField.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited;
  FFieldPrefixies := TgaSQLTokenList.CreateMirror(AOwnerStatement, Self);
  SetIsExpression(False);
  OwnerStatement.AddField(Self);
end;

destructor TgaSQLField.Destroy;
begin
  if Assigned(OwnerStatement) then
    OwnerStatement.RemoveField(Self);
  FFieldAlias.Free;
  FFieldName.Free;
  FFieldPrefixies.Free;
  inherited Destroy;
end;

procedure TgaSQLField.ExecuteTokenAdded(Sender: TObject; AToken: 
        TgaSQLTokenObj);
begin
  inherited;
  { Field is written in SQL as follows:
    [table alias][.]FieldName[whitespace][FieldAlias],}
  if AToken.TokenType = stComment then
    Exit;
  if not IsExpression then
  begin
    case FieldParseState of
      fepsNoSymbolsParsed:
        ParsePreFieldExpression(AToken);
      fepsFieldExpression:
        ParseFieldExpression(AToken);
      fepsFieldAlias:
        ParseFieldAlias(AToken);
      fepsFieldAliasParsed:
        if not (AToken.TokenType in [stDelimitier, stComma]) then
          StartExpressionParse;
      fepsFieldParseFinished:
        raise Exception.Create(SerrFieldParseFinised);
      { fepsExpression:
        Shouldn't get here as IsExpression should be true then }
      else
        raise Exception.CreateFmt(SerrUnecpectedFieldParseState,
          [GetEnumName(TypeInfo(TFieldExprParseState), Ord(FieldParseState))]);
    end;
    if (AToken.TokenType = stComma) then
      ParseComplete := True;
  end;
end;

function TgaSQLField.GetFieldAlias: string;
begin
  if Assigned(FFieldAlias) then
    Result := GetTokenObjAsString(FFieldAlias.TokenObj)
  else
    Result := EmptyStr;
end;

function TgaSQLField.GetFieldName: string;
begin
  if Assigned(FFieldName) then
    Result := GetTokenObjAsString(FFieldName.TokenObj)
  else
    Result := EmptyStr;
end;

function TgaSQLField.GetFieldPrefix: string;
begin
  Result := FFieldPrefixies.AsString;
end;

procedure TgaSQLField.InternalSetParseComplete;
begin
  if FFieldParseState = fepsFieldExpression then
    TerminateFieldExpressionParse;
  inherited;
  FFieldParseState := fepsFieldParseFinished;
end;

procedure TgaSQLField.ParseFieldAlias(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stDelimitier, stComma:
      {no special processing};
    stSymbol, stQuotedSymbol, stString:
      if not AToken.TokenSymbolIs('AS') then
      begin
        FFieldAlias := GetBookmark;
        FFieldParseState := fepsFieldAliasParsed;
      end;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.ParseFieldExpression(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
      ;
    stDelimitier, stComma:
      TerminateFieldExpressionParse;
    stOther:
      if AToken.TokenString <> '*' then
        StartExpressionParse;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.ParsePreFieldExpression(AToken: TgaSQLTokenObj);
begin
  case AToken.TokenType of
    stSymbol, stQuotedSymbol, stPeriod:
    begin
      FFieldPrefixies.SetStartPos(Self, True);
      FFieldParseState := fepsFieldExpression;
    end;
    stDelimitier:
        { skip leading whitespaces };
    stComma:
      IsInvalid := True;
    stOther:
      if AToken.TokenString = '*' then
      begin
        FFieldName := GetBookmark;
        FFieldParseState := fepsFieldAliasParsed;
      end else
        StartExpressionParse;
    else
      StartExpressionParse;
  end;
end;

procedure TgaSQLField.SetFieldAlias(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
  tmpToken: TgaSQLTokenObj;
begin
  if FieldAlias <> Value then
  begin
    CheckModifyAllowed;
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count > 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Field alias', 1, tmpTokenList.Count]);
      tmpTokenList.First;
      if Assigned(FFieldAlias) then
        FFieldAlias.TokenObj := tmpTokenList.CurrentItem
      else begin
        if IsExpression then
          Last
        else
          GotoBookmark(FFieldName);
        tmpToken := TgaSQLTokenObj.Create;
        tmpToken.SetTokenInfo(' ', stDelimitier, False, #0);
        InsertAfterCurrent(tmpToken, True);
        InsertAfterCurrent(tmpTokenList.CurrentItem, True);
        FFieldAlias := GetBookmark;
      end;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.SetFieldName(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
begin
  if FieldName <> Value then
  begin
    CheckModifyAllowed;
    if IsExpression then
      raise Exception.Create(SerrFieldAttrCantBeChangedInExpression);
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(Value, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      if tmpTokenList.Count <> 1 then
        raise Exception.CreateFmt(SerrWrongTokenCountInArg, ['Field name', 1, tmpTokenList.Count]);
      tmpTokenList.First;
      FFieldName.TokenObj := tmpTokenList.CurrentItem;
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.SetFieldPrefix(const Value: string);
var
  tmpStr: string;
  tmpTokenList: TgaSQLTokenHolderList;
begin
  if Value[Length(Value)] = '.' then
    tmpStr := Value
  else
    tmpStr := Value + '.';
  if (FieldPrefix <> tmpStr) then
  begin
    CheckModifyAllowed;
    if IsExpression then
      raise Exception.Create(SerrFieldAttrCantBeChangedInExpression);
    tmpTokenList := TgaSQLTokenHolderList.Create(nil);
    try
      ParseStringToTokens(tmpStr, tmpTokenList);
      TrimTokenList(tmpTokenList, True);
      FieldPrefixies.CopyListContest(tmpTokenList)
    finally
      tmpTokenList.Free;
    end;
  end;
end;

procedure TgaSQLField.StartExpressionParse;
begin
  FreeAndNil(FFieldName);
  FreeAndNil(FFieldAlias);
  FreeAndNil(FFieldPrefixies);
  FIsExpression := True;
  FFieldParseState := fepsExpression;
  First;
  while not Eof do
  begin
    ParseExpression(CurrentItem);
    Next;
  end;
end;

procedure TgaSQLField.TerminateFieldExpressionParse;
begin
  if FFieldParseState <> fepsFieldExpression then
    Exit;
  FFieldPrefixies.Last;
  // last one should be the current comma or delimitier
  FFieldPrefixies.Previous;
  if FFieldPrefixies.CurrentItem.TokenType <> stPeriod then
  begin
    Locate(FFieldPrefixies.CurrentItem);
    FFieldName := GetBookmark;
    FFieldPrefixies.Previous;
  end;
  FFieldPrefixies.SetEndPos(FFieldPrefixies, True);
  FFieldParseState := fepsFieldAlias;
  { #ToDo2 Should there be an property do determ whete field alias is
            allowed or not }
end;

{
*************************** TgaSQLStatementPartList ****************************
}
constructor TgaSQLStatementPartList.Create(AOwnerStatement: 
        TgaCustomSQLStatement; AStatementPartType: TgaSQLStatementPartClass);
begin
  inherited Create;
  FTokenList := TgaSQLTokenList.CreateMirror(AOwnerStatement, 
          AOwnerStatement.CurrentSQL);
  FTokenList.SetStartPos(AOwnerStatement.CurrentSQL, True);
  FStatementPartType := AStatementPartType;
  FOwnerStatement := AOwnerStatement;
end;

destructor TgaSQLStatementPartList.Destroy;
begin
  FreeAndNil(FTokenList);
  inherited Destroy;
end;

procedure TgaSQLStatementPartList.Clear;
begin
  FCurrentPart := nil;
  inherited Clear;
  if Assigned(TokenList) then
    TokenList.Clear;
end;

procedure TgaSQLStatementPartList.ExecuteTokenAdded(Sender: TObject; AToken: 
        TgaSQLTokenObj);
begin
  CurrentPart.ExecuteTokenAdded(Sender, AToken);
  if CurrentPart.ParseComplete then
    FCurrentPart := nil;
  TokenList.Locate(AToken);
end;

function TgaSQLStatementPartList.GetAsString: string;
begin
  Result := inherited GetAsString;
end;

function TgaSQLStatementPartList.GetCurrentPart: TgaSQLStatementPart;
begin
  if not Assigned(FCurrentPart) then
  begin
    FCurrentPart := StatementPartType.Create(OwnerStatement);
    Add(FCurrentPart);
  end;
  Result := FCurrentPart;
end;

function TgaSQLStatementPartList.GetParseComplete: Boolean;
begin
  Result := FParseComplete;
end;

procedure TgaSQLStatementPartList.InternalSetParseComplete;
begin
  if Assigned(FCurrentPart) then
    FCurrentPart.ParseComplete := True;
  FParseComplete := True;
  TokenList.SetEndPos(TokenList, True);
  TokenList.ActiveDataShare := False;
end;

procedure TgaSQLStatementPartList.SetAsString(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
  AItem: TgaSQLTokenObj;
begin
  tmpTokenList := TgaSQLTokenHolderList.Create(nil);
  try
    ParseStringToTokens(Value, tmpTokenList);
    TrimTokenList(tmpTokenList, True, [stEnd]);
    if tmpTokenList.Count = 0 then
      tmpTokenList.AddToken(TgaSQLTokenObj.CreatePlaceHolder);
    Clear;
    tmpTokenList.First;
    while not tmpTokenList.Eof do
    begin
      AItem := tmpTokenList.CurrentItem;
      TokenList.Add(AItem);
      // #Todo2 this can be optimized
      OwnerStatement.CurrentSQL.Locate(AItem);
      ExecuteTokenAdded(Self, AItem);
      tmpTokenList.Next;
    end;
  finally
    tmpTokenList.Free;
  end;
end;

procedure TgaSQLStatementPartList.SetParseComplete(const Value: Boolean);
begin
  if Value then
    InternalSetParseComplete;
end;

{
******************************* TgaSQLFieldList ********************************
}
constructor TgaSQLFieldList.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerSTatement, TgaSQLField);
end;

{
******************************* TgaSQLTableList ********************************
}
constructor TgaSQLTableList.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerSTatement, TgaSQLTable);
end;

{
****************************** TgaSQLOrderByList *******************************
}
function TgaSQLOrderByList.GetAsString: string;
begin
  Result := EmptyStr;
  if TokenList.IsEmpty then
    Exit;
  TokenList.First;
  if TokenList.CurrentItem.TokenSymbolIs('ORDER') then
  begin
    TokenList.Next;
    while (not TokenList.Eof) and (TokenList.CurrentItem.TokenType = stDelimitier) do
      TokenList.Next;
    if TokenList.CurrentItem.TokenSymbolIs('BY') then
    begin
      TokenList.Next;
      while (not TokenList.Eof) and (TokenList.CurrentItem.TokenType = stDelimitier) do
        TokenList.Next;
    end;
  end;
  while not TokenList.Eof do
  begin
    Result := Result + TokenList.CurrentItem.TokenAsString;
    TokenList.Next;
  end;
end;

procedure TgaSQLOrderByList.SetAsString(const Value: string);
var
  tmpStr: string;
  tmpToken: TgaSQLTokenObj;
begin
  if (Trim(Value) = '') or (SameText('order by ', Copy(Value, 1, 9))) then
    tmpStr := Value
  else
    tmpStr := 'order by ' + Value;
  inherited SetAsString(tmpStr);
  TokenList.Last;
  if TokenList.CurrentItem.TokenType <> stDelimitier then
  begin
    tmpToken := TgaSQLTokenObj.Create;
    tmpToken.SetTokenInfo(#13#10, stDelimitier, False, #0);
    TokenList.Add(tmpToken);
    ExecuteTokenAdded(Self, tmpToken);
  end;
end;

{
***************************** TgaSQLStatementPart ******************************
}
constructor TgaSQLStatementPart.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  if not Assigned(AOwnerStatement) then
    raise Exception.Create(SerrStmPartWithoutStm);
  inherited CreateMirror(AOwnerStatement, AOwnerStatement.CurrentSQL);
end;

procedure TgaSQLStatementPart.CheckModifyAllowed;
begin
  if not Assigned(OwnerStatement) then
    raise Exception.Create(SerrNowOwnerParser)
end;

function TgaSQLStatementPart.GetAsString: string;
begin
  Result := inherited GetAsString;
end;

function TgaSQLStatementPart.GetParseComplete: Boolean;
begin
  Result := InternalGetParseComplete;
end;

function TgaSQLStatementPart.InternalGetParseComplete: Boolean;
begin
  Result := FParseComplete;
end;

procedure TgaSQLStatementPart.InternalSetParseComplete;
begin
  FParseComplete := True;
  SetEndPos(Self, True);
  ActiveDataShare := False;
end;

procedure TgaSQLStatementPart.SetAsString(const Value: string);
var
  tmpTokenList: TgaSQLTokenHolderList;
begin
  CheckModifyAllowed;
  tmpTokenList := TgaSQLTokenHolderList.Create(nil);
  try
    ParseStringToTokens(Value, tmpTokenList);
    TrimTokenList(tmpTokenList, True, [stEnd]);
    if tmpTokenList.Count = 0 then
      tmpTokenList.AddToken(TgaSQLTokenObj.CreatePlaceHolder);
    CopyListContest(tmpTokenList);
  finally
    tmpTokenList.Free;
  end;
end;

procedure TgaSQLStatementPart.SetIsInvalid(const Value: Boolean);
begin
  FIsInvalid := Value;
end;

procedure TgaSQLStatementPart.SetParseComplete(const Value: Boolean);
begin
  if Value then
    InternalSetParseComplete;
end;

{
******************************* TgaSQLExpression *******************************
}
constructor TgaSQLExpression.Create(AOwnerStatement: TgaCustomSQLStatement);
begin
  inherited Create(AOwnerStatement);
  FisExpression := True;
  FIsCorrect := True;
end;

procedure TgaSQLExpression.AddParen(AToken: TgaSQLTokenObj);
begin
  if FParenCount = 0 then
  begin
    FOldOnStatementAdd := OwnerStatement.OnTokenAdded;
    OwnerStatement.OnTokenAdded := ExecuteTokenAdded;
  end;
  Inc(FParenCount);
end;

procedure TgaSQLExpression.ExecuteTokenAdded(Sender: TObject; AToken: 
        TgaSQLTokenObj);
begin
  inherited ;
  if AToken.TokenType = stComment then
    Exit;
  if IsExpression then
    ParseExpression(AToken);
end;

function TgaSQLExpression.GetCanParseEnd: Boolean;
begin
  Result := ParenCount = 0;
end;

function TgaSQLExpression.GetExpression: string;
begin
  if IsExpression then
    Result := AsString
  else
    Result := EmptyStr;
end;

procedure TgaSQLExpression.ParseExpression(AToken: TgaSQLTokenObj);
begin
  // #Todo3 Expression parsing not implemented... yet
  case AToken.TokenType of
    stLParen:
      AddParen(AToken);
    stRParen:
      RemoveParen(AToken);
  end;
  if ParenCount = 0 then
    case AToken.TokenType of
      stLParen, stRParen: ;// Do not parse as parsed already
      stComma:
        ParseComplete := True;
    end;
end;

procedure TgaSQLExpression.RemoveParen(AToken: TgaSQLTokenObj);
begin
  Dec(FParenCount);
  if ParenCount = 0 then
    OwnerStatement.OnTokenAdded := FOldOnStatementAdd;
  if ParenCount < 0 then
    FIsCorrect := False;
end;

procedure TgaSQLExpression.SetIsExpression(const Value: Boolean);
begin
  FIsExpression := Value;
end;

{ TgaJoinOnClause }

{
******************************* TgaJoinOnClause ********************************
}
function TgaJoinOnClause.GetAsString: string;
begin
  Result := EmptyStr;
  if IsEmpty then
    Exit;
  First;
  if CurrentItem.TokenSymbolIs('ON') then
  begin
    Next;
    while (not Eof) and (CurrentItem.TokenType = stDelimitier) do
      Next;
  end;
  while not Eof do
  begin
    Result := Result + CurrentItem.TokenAsString;
    Next;
  end;
end;

procedure TgaJoinOnClause.SetAsString(const Value: string);
var
  tmpStr: string;
  tmpToken: TgaSQLTokenObj;
begin
  if (Trim(Value) = '') or (SameText('on ', Copy(Value, 1, 3))) then
    tmpStr := Value
  else
    tmpStr := 'on ' + Value;
  inherited SetAsString(tmpStr);
  Last;
  if CurrentItem.TokenType <> stDelimitier then
  begin
    tmpToken := TgaSQLTokenObj.Create;
    tmpToken.SetTokenInfo(#13#10, stDelimitier, False, #0);
    Add(tmpToken);
    ExecuteTokenAdded(Self, tmpToken);
  end;
end;

end.
